{-----------------------------------------------------------------------------
 Unit Name: Synth
 Author:    Sebastian Htter
 Date:      2006-10-29
 Purpose:   Simple Synthesizer, 'BlitzTracker'
            Parts taken from a Tutorial by Nico Michaelis

 History:   2006-10-29 initial release
 
------------------------------------------------------------------------------
 Copyright Notice: If you copy or modify this file, this block
   _must_ be preserved! If you are using this code in your projects,
   I would like to see a word of thanks in the About-Box or a similar place.
-----------------------------------------------------------------------------}
unit Synth;

interface

{$DEFINE SUPPORT_VER0}
{$DEFINE SUPPORT_VER1}
{$DEFINE SUPPORT_VER2}

uses Windows,
     fmod, fmodtypes, fmoderrors, fmod_synthutils,
     engine_rtf,
     SynthUtils,
     SynthInstruments,
     Classes;


const
  SamplingRate = 44100;
                             // ++---- category
  //Synthethic               // ||++-- instrument
  Synth_Instr_Sine          =  $0000;
  Synth_Instr_Triangle      =  $0001;
  Synth_Instr_Sawtooth      =  $0002;
  Synth_Instr_Pulse         =  $0003;
  Synth_Instr_Noise         =  $0004;
  //Keys
  Synth_Instr_Piano         =  $0100;
  Synth_Instr_Clavinet1     =  $0101;
  Synth_Instr_Clavinet2     =  $0102;
  Synth_Instr_EOrgan        =  $0103;
  //Woodwind
  Synth_Instr_Flute         =  $0200;
  Synth_Instr_Trumpet       =  $0201;
  //Percussion
  Synth_Instr_Bell          =  $0300;
  Synth_Instr_Drum          =  $0301;
  Synth_Instr_SynthDrum     =  $0302;
  Synth_Instr_Snare         =  $0303;
  Synth_Instr_Hat           =  $0304;

type
  EFileTypeError = class(EStreamError);

  String25= string[25];
  TEventType = (seNone, seOn, seOff, sePitch, sePan, seRepeat, seText, seVolume);
  TSynthEvent = packed record
    TimeBefore: Word;
    case EventType:TEventType of
      seNone: ();
      seOn : ();
      seOff: ();
      sePitch: (Freq:single);
      sePan: (Panning:single);
      seRepeat: (last,times:word);
      seText: (Text:String25);
      seVolume: (Volume:single);
    end;
  TSynthTrack = record
                  TrackName: String25;
                  InstrumentID:Word;
                  EventCount:integer;
                  Events:array of TSynthEvent;
                end;
  TSynthSong = record
                  Creator,
                  Title,
                  Composer:String25;
                  TrackCount:integer;
                  Tracks: array of TSynthTrack;
               end;

  TCompInstr = Record
                 Started:integer;
                 Duration,
                 Freq,
                 Pan,
                 Volume:single;
                 InstrID:word;
               end;

  TCompTrack = record
                 Instr:TCompInstr;
                 Events:array of TSynthEvent;
               end;

  TCompSong = array of TCompTrack;

procedure SynthInitSong(var Song:TSynthSong);
function SynthAddTrack(var Song:TSynthSong):integer;
procedure SynthRemoveTrack(var Song:TSynthSong; Tracknum:integer);
function SynthInsertEvent(var Track:TSynthTrack; EventNum:integer):integer;
procedure SynthRemoveEvent(var Track:TSynthTrack; EventNum:integer);

function SynthCompileSong(Song:TSynthSong):TCompSong;
function SynthGenerateSong(Instr: word; Freq: single; Duration: integer):TCompSong;
function SynthSongLength(Song:TSynthSong):integer;
function SynthCompiledSongLength(Song:TCompSong):integer;
procedure SynthPlaySong(Song:TCompSong; StartTime:integer); overload;
procedure SynthPlaySong(Song:TSynthSong; StartTime:integer); overload;
function SynthGetPos:integer;
procedure SynthStopSong;

procedure SynthSaveToStream(Song:TSynthSong; Stream:TStream);
procedure SynthSaveToFile(Song:TSynthSong; FN:String);

procedure SynthLoadFromFile(var Song:TSynthSong; FN:String; Version: byte=$FF);


{$IFDEF SUPPORT_VER2}procedure SynthLoadFromStream2(var Song:TSynthSong; Stream:TStream); {$ENDIF}
{$IFDEF SUPPORT_VER1}procedure SynthLoadFromStream1(var Song:TSynthSong; Stream:TStream); {$ENDIF}
{$IFDEF SUPPORT_VER0}procedure SynthLoadFromStream0(var Song:TSynthSong; Stream:TStream); {$ENDIF}

procedure SynthListInstruments(List:TStrings);
function EventToStr(Ev:TSynthEvent):string;
function SynthInstrumentName(ID:word):string;

implementation

uses SysUtils;

procedure TerminateString(var S:String25);
var l:integer;
begin
  SetLength(S,25);
  for l:= 1 to length(S) do
    if S[l]=#0 then begin SetLength(S,l-1); break; end;
end;

const
  FileSignature: byte = $80 or 2; 
  InstrNames : array[0..15] of record Name:string; ID:word; end =
    (
      (Name:'Sine Wave'    ; ID:Synth_Instr_Sine      ),
      (Name:'Triangle Wave'; ID:Synth_Instr_Triangle  ),
      (Name:'Sawtooth Wave'; ID:Synth_Instr_Sawtooth  ),
      (Name:'Pulse Wave'   ; ID:Synth_Instr_Pulse     ),
      (Name:'Noise'        ; ID:Synth_Instr_Noise     ),
      //Keys
      (Name:'Piano'        ; ID:Synth_Instr_Piano     ),
      (Name:'Clavinet1'    ; ID:Synth_Instr_Clavinet1 ),
      (Name:'Clavinet2'    ; ID:Synth_Instr_Clavinet2 ),
      (Name:'EOrgan'       ; ID:Synth_Instr_EOrgan    ),
      //Woodwind
      (Name:'Flute'        ; ID:Synth_Instr_Flute     ),
      (Name:'Trumpet'      ; ID:Synth_Instr_Trumpet   ),
      //Percussion
      (Name:'Bell'         ; ID:Synth_Instr_Bell      ),
      (Name:'Drum'         ; ID:Synth_Instr_Drum      ),
      (Name:'SynthBase'    ; ID:Synth_Instr_SynthDrum ),
      (Name:'Snare'        ; ID:Synth_Instr_Snare     ),
      (Name:'Hat'          ; ID:Synth_Instr_Hat       )
    );


function EventToStr(Ev:TSynthEvent):string;
begin
  Result:= IntToStr(Ev.TimeBefore)+':';
  case Ev.EventType of
    seNone : Result:= Result+'Pause';
    seOn   : Result:= Result+'On';
    seOff  : Result:= Result+'Off';
    sePitch: Result:= Result+'Pitch='+FloatToStrF(Ev.Freq,ffFixed,15,2);
    sePan  : Result:= Result+'Panning='+FloatToStrF(Ev.Panning,ffFixed,15,2);
    seRepeat: Result:= Result+'Repeat last '+IntToStr(Ev.last)+' events '+IntToStr(ev.times)+' times';
    seText : Result:= '---'+Ev.Text;
    seVolume: Result:= Result+'Volume='+FloatToStrF(Ev.Panning,ffFixed,15,2);
  end;
end;

procedure SynthListInstruments(List:TStrings);
  procedure Append(S:string; I:integer);
  begin
    List.AddObject(S+#9'($'+IntToHex(i,4)+')',Pointer(i));
  end;
var i:integer;
begin
  List.Clear;
  for i:= 0 to high(InstrNames) do
    Append(InstrNames[i].Name, InstrNames[i].ID);
end;

function SynthInstrumentName(ID:word):string;
var i:integer;
begin
  Result:= '';
  for i:= 0 to high(InstrNames) do
    if InstrNames[i].ID = ID then begin
      Result:= InstrNames[i].Name;
      exit;
    end;
end;


procedure SynthInitSong(var Song:TSynthSong);
begin
  FillChar(Song, Sizeof(Song),0);
end;

function SynthAddTrack(var Song:TSynthSong):integer;
begin
  Result:= song.TrackCount;
  inc(song.TrackCount);
  SetLength(Song.Tracks,Song.TrackCount);
  FillChar(Song.Tracks[Result],sizeof(TSynthTrack),0);
  Song.Tracks[Result].TrackName:= 'Track '+IntToStr(Song.TrackCount);
end;

procedure SynthRemoveTrack(var Song:TSynthSong; Tracknum:integer);
var i:integer;
begin
  for i:= Song.TrackCount-1 downto Tracknum+1 do
    Song.Tracks[i-1]:= Song.Tracks[i];
  dec(song.TrackCount);
  SetLength(Song.Tracks,Song.TrackCount);
end;

procedure SynthRemoveEvent(var Track:TSynthTrack; EventNum:integer);
var
  i: Integer;
begin
  for i := EventNum to Track.EventCount-2 do begin //von "pos" bis vorletztes Element durchgehen...
    Track.Events[i] := Track.Events[i+1];
  end;
  dec(Track.EventCount);
  SetLength(Track.Events,Track.EventCount);
end;

function SynthInsertEvent(var Track:TSynthTrack; EventNum:integer):integer;
var i:integer;
begin
  inc(Track.EventCount);
  SetLength(Track.Events,Track.EventCount);
  if EventNum>Track.EventCount-1 then
    EventNum:= -1;
  if (EventNum>-1) and (Track.EventCount>1) then begin
    for i:= Track.EventCount-1 downto EventNum do
      Track.Events[i]:= Track.Events[i-1];
  end else
    EventNum:= Track.EventCount-1;
  FillChar(Track.Events[EventNum],Sizeof(TSynthEvent),0);
  Result:= EventNum;
end;

function SynthCompileSong(Song:TSynthSong):TCompSong;
var tr:integer;

  procedure CompileTrack(A:TSynthTrack; var B: TCompTrack);
  var time,
      pos:integer;
    procedure Grow(var X:TCompTrack);
    begin
      SetLength(X.Events,trunc((length(X.Events)+1)*1.7));
    end;

    procedure ProcessSub(Ev1,Ev2:integer);
    var e,n:integer;
    begin
      for e:= Ev1 to Ev2 do begin
        inc(time,A.Events[e].TimeBefore);
        case A.Events[e].EventType of
          seNone,
          seText: continue;  //drop them....
          seRepeat : for n:= 1 to A.Events[e].Times do begin
                       ProcessSub(e-A.Events[e].last,e-1);
                     end;
        else
          B.Events[pos]:= A.Events[e];
          B.Events[pos].TimeBefore:= time;
          inc(pos);
        end;
        if pos>=high(B.Events) then
          Grow(B);
      end;
    end;
  begin
    SetLength(B.Events,A.EventCount);
    B.Instr.InstrID:= A.InstrumentID;
    B.Instr.Pan:= 0.5;
    B.Instr.Started:= -1;
    B.Instr.Freq:= 440;
    B.Instr.Volume:= 1.0;
    time:= 0;
    pos:= 0;
    ProcessSub(0,A.EventCount-1);
    SetLength(B.Events,pos);
  end;

begin
  SetLength(Result,Song.TrackCount);
  for tr:= 0 to Song.TrackCount-1 do begin
    CompileTrack(Song.Tracks[tr],Result[tr]);
  end;
end;

function SynthGenerateSong(Instr: word; Freq: single; Duration: integer):TCompSong;
begin
  SetLength(Result,1);
  SetLength(Result[0].Events,2);
  Result[0].Instr.InstrID:= Instr;
  Result[0].Instr.Pan:= 0.5;
  Result[0].Instr.Started:= -1;
  Result[0].Instr.Freq:= Freq;
  Result[0].Instr.Volume:= 1.0;
  Result[0].Events[0].EventType:= seOn;
  Result[0].Events[1].EventType:= seOff;
  Result[0].Events[1].TimeBefore:= Duration;
end;

function SynthSongLength(Song:TSynthSong):integer;
var
  t, l: integer;
  A: TSynthTrack;
    procedure ProcessSub(Ev1,Ev2:integer);
    var e,n:integer;
    begin
      for e:= Ev1 to Ev2 do begin
        inc(l,A.Events[e].TimeBefore);
        case A.Events[e].EventType of
          seRepeat : for n:= 1 to A.Events[e].Times do begin
                       ProcessSub(e-A.Events[e].last,e-1);
                     end;
        end;
      end;
    end;
begin
  Result:= 0;
  for t:= 0 to Song.TrackCount - 1 do begin
    l:= 0;
    A:= Song.Tracks[t];
    ProcessSub(0,A.EventCount-1);
    if l>Result then
      Result:= l;
  end;
end;

function SynthCompiledSongLength(Song:TCompSong):integer;
var t, l: integer;
begin
  Result:= 0;
  for t:= 0 to high(Song) do begin
    l:= Song[t].Events[high(Song[t].Events)].TimeBefore;
    if l > Result then
      Result:= l;
  end;
end;

var
  SoundActive : Boolean;
  Stream: PFSoundStream;
  Channel : Integer;
  Pos : Integer;

  Playing:TCompSong;

function StreamCallback(Stream: PFSoundStream; Buff: Pointer; Len, Param: Integer): ByteBool; stdcall;
var
  Count,
  tr: Integer;
  Buffer: PChar;
  Value,
  ValueL,ValueR : Single;
  TimeAft:single;

  procedure AdjustDuration(var I:TCompInstr);
  var SustainTime : Single;
  begin
    i.Duration := i.Duration/1000;
    case I.InstrID of
      Synth_Instr_Piano,
      Synth_Instr_Clavinet1,
      Synth_Instr_Clavinet2: SustainTime := i.Duration
                                            - Klavier_Waves_Attack/i.Freq
                                            - Klavier_Waves_End/i.Freq;
      Synth_Instr_EOrgan: SustainTime := i.Duration
                                         - EOrgel_Waves_Attack/i.Freq
                                         - EOrgel_Waves_Decay/i.Freq
                                         - EOrgel_Waves_End/i.Freq;
      Synth_Instr_Flute,Synth_Instr_Trumpet:
                     SustainTime := i.Duration
                       - Floete_Waves_Attack/i.Freq
                       - Floete_Waves_End/i.Freq;
      Synth_Instr_Bell:
                     SustainTime := i.Duration
                       - Glocke_Waves_Attack/i.Freq
                       - Glocke_Waves_End/i.Freq;
      Synth_Instr_Drum,
      Synth_Instr_SynthDrum,
      Synth_Instr_Snare:
                     SustainTime := i.Duration
                       - Drum_Waves_Attack/i.Freq
                       - Drum_Waves_Release/i.Freq;
    else
      SustainTime:= 0;
    end;
    if SustainTime < 0 then
      i.Duration := i.Duration - SustainTime;
  end;


  procedure EventProcessor;
  var i,j,k:integer;
  begin
    for i:= 0 to high(Playing) do begin
      for j:= 0 to high(Playing[i].Events) do begin
        with Playing[i].Events[j] do begin
          if (TimeBefore<$FFFF) and (TimeBefore<=TimeAft) then begin
            case EventType of
              seOn   :begin
                        Playing[i].Instr.Started:= pos;
                        for k:= j+1 to high(Playing[i].Events) do
                          if Playing[i].Events[k].EventType=seOff then begin
                            Playing[i].Instr.Duration:= Playing[i].Events[k].TimeBefore-TimeBefore;
                            AdjustDuration(Playing[i].Instr);
                            break;
                          end;
                      end;  
              seOff  :Playing[i].Instr.Started:= -1;
              sePitch:Playing[i].Instr.Freq:= Freq;
              sePan  :Playing[i].Instr.Pan:= Panning;
              seVolume:Playing[i].Instr.Volume:= Volume;
            end;
            TimeBefore:= $FFFF;  // dont play again
          end;
        end;
      end;
    end;
  end;

  function PlayInstr(I:integer; F,T,D:single):single;
  var x,y:single;
    function Sgn(x : Single) : Single;
    begin
      if x < 0 then
        Result := -1.0
      else
        Result := +1.0;
    end;

  begin
    y:= 1/SamplingRate * T;
    X:= 1/SamplingRate * F * T;
    case I of
      Synth_Instr_Sine:
           Result:= WaveSine(X);
      Synth_Instr_Triangle:
           Result:= WaveTriangle(X);
      Synth_Instr_Sawtooth:
           Result:= WaveSawtooth(X);
      Synth_Instr_Pulse:
           Result:= WavePulse(X,f/5000);
      Synth_Instr_Noise:
           Result:= (Random - 0.5) * 2;

      Synth_Instr_Piano:
           Result:= PlayPiano(Y,f,D);
      Synth_Instr_Clavinet1:
           Result:= PlayClavinetA(Y,f,D);
      Synth_Instr_Clavinet2:
           Result:= PlayClavinetB(Y,f,D);
      Synth_Instr_EOrgan:
           Result:= PlayEOrgan(Y,f,D);

      Synth_Instr_Flute:
           Result:= PlayFlute(Y,f,D);
      Synth_Instr_Trumpet:
           Result:= PlayTrumpet(Y,f,D);

      Synth_Instr_Bell:
           Result:= PlayBell(Y,f,D);
      Synth_Instr_Drum:
           Result:= PlayDrum(Y,f,D);
      Synth_Instr_SynthDrum:
           Result:= PlaySynthDrum(Y,f,D);
      Synth_Instr_Snare:
           Result:= PlaySnare(Y,f,D);
      Synth_Instr_Hat:
           Result:= PlayHat(Y,f,D);
    else
      Result:= 0;
    end;
  end;


begin
  Buffer := PChar(Buff);
  Count := 0;

  while Count < (Len shr 2) do		{ >>2 = 16bit stereo (4 bytes per sample) }
  begin
    TimeAft:= Pos/SamplingRate*1000;
    EventProcessor;

    Value:= 0;
    valueL:=0;
    valueR:=0;
    for tr:= 0 to high(Playing) do begin
      if playing[tr].Instr.Started<0 then continue;
      Value := Value + PlayInstr(playing[tr].Instr.InstrID,
                                 playing[tr].Instr.Freq,
                                 pos-playing[tr].Instr.Started,
                                 playing[tr].Instr.Duration)
                     * playing[tr].Instr.Volume;
      valueL:= Value * playing[tr].Instr.Pan;
      valueR:= Value * (1-playing[tr].Instr.Pan);
    end;


    PSmallInt(Buffer)^ := Trunc(ValueL * (32767.0/2.0));    { Left channel }
    Inc(Buffer, 2);
    PSmallInt(Buffer)^ := Trunc(ValueR * (32767.0/2.0));   { Right channel }
    Inc(Buffer, 2);

    Inc(Count);
    Inc(Pos);
  end;

  Result := True;
end;

procedure SynthPlaySong(Song:TCompSong; StartTime:integer);
begin
  if SoundActive then
    SynthStopSong;
  Pos := StartTime * SamplingRate div 1000;

  Playing:= Song;

  SoundActive := OpenFMOD;
  if SoundActive then
  begin
    if not FSOUND_Init(SamplingRate, 16, 0) then
    begin
      LogFile.AddWarning('->FormCreate',
        Format('Bei der Initialisierung von FMOD ist folgender Fehler aufgetreten: %s',
               [FMOD_ErrorString(FSOUND_GetError)]));
      SoundActive := false;
      Exit;
    end;
                                                 {elem *16bit* stereo}
    Stream := FSOUND_Stream_Create(StreamCallback, 2500 * 2 * 2, FSOUND_NORMAL or FSOUND_16BITS or FSOUND_STEREO, SamplingRate, 12345);
    if Stream = nil then
    begin
      LogFile.AddWarning('->FormCreate',
        Format('Bei der des Sound-Streams ist folgender Fehler aufgetreten: %s',
               [FMOD_ErrorString(FSOUND_GetError)]));
      SoundActive := false;
      Exit;
    end;
    Channel := FSOUND_Stream_Play(FSOUND_FREE, Stream);
  end
  else
    LogFile.AddWarning('->FormCreate', 'Soundausgabe nicht mglich!');
end;

procedure SynthPlaySong(Song:TSynthSong; StartTime:integer);
begin
  SynthPlaySong(SynthCompileSong(Song),StartTime);
end;

function SynthGetPos:integer;
begin
  if SoundActive then
    Result:= trunc(pos/SamplingRate*1000)
  else
    Result:= -1;
end;

procedure SynthStopSong;
begin
  if SoundActive then
    FSOUND_Stream_Close(Stream);
  SoundActive:= false;
  CloseFMOD;
end;

procedure SynthSaveToStream(Song:TSynthSong; Stream:TStream);
var i,j:integer;
begin
  Stream.Write(FileSignature,1);
  Stream.Write(song.Creator,Sizeof(song.Creator));
  Stream.Write(song.Title,Sizeof(song.Title));
  Stream.Write(song.Composer,Sizeof(song.Composer));
  Stream.Write(song.TrackCount,Sizeof(Integer));
  for i:= 0 to Song.TrackCount-1 do begin
    Stream.Write(song.Tracks[i].TrackName,Sizeof(song.Tracks[j].TrackName));
    Stream.Write(song.Tracks[i].InstrumentID,Sizeof(song.Tracks[j].InstrumentID));
    Stream.Write(song.Tracks[i].EventCount,Sizeof(song.Tracks[j].EventCount));
    for j:= 0 to Song.Tracks[i].EventCount-1 do begin
      Stream.Write(Song.Tracks[i].Events[j].TimeBefore,sizeof(Word));
      Stream.Write(Song.Tracks[i].Events[j].EventType,sizeof(TEventType));
      case Song.Tracks[i].Events[j].EventType  of
        seNone: ;
        seOn : ;
        seOff: ;
        sePitch: Stream.Write(Song.Tracks[i].Events[j].Freq,Sizeof(Single));
        sePan: Stream.Write(Song.Tracks[i].Events[j].Panning,Sizeof(Single));
        seRepeat: begin
          Stream.Write(Song.Tracks[i].Events[j].Last,Sizeof(Word));
          Stream.Write(Song.Tracks[i].Events[j].Times,Sizeof(Word));
        end;
        seText: Stream.Write(Song.Tracks[i].Events[j].Text[1],Sizeof(String25));
        seVolume: Stream.Write(Song.Tracks[i].Events[j].Volume,Sizeof(Single));
      end;
    end;
 end;
end;

procedure SynthSaveToFile(Song:TSynthSong; FN:String);
var f:TFileStream;
begin
  f:= TFileStream.Create(FN,fmOpenWrite or fmCreate);
  try
    SynthSaveToStream(Song,f);
  finally
    f.Free;
  end;
end;

procedure SynthLoadFromFile(var Song:TSynthSong; FN:String; Version: byte);
var f:TFileStream;
    b:byte; 
begin
  f:= TFileStream.Create(FN,fmOpenRead);
  try
    if Version=$FF then begin
      f.Read(b,1);
      if b and $80 = $80 then
        Version:= b and $7F;
    end;
    case Version of
{$IFDEF SUPPORT_VER0}  0: SynthLoadFromStream0(Song,f);  {$ENDIF}
{$IFDEF SUPPORT_VER1}  1: SynthLoadFromStream1(Song,f);  {$ENDIF}
{$IFDEF SUPPORT_VER2}  2: SynthLoadFromStream2(Song,f);  {$ENDIF}
      $FF: raise EFileTypeError.Create('Cannot autodetect file version. Please select manually.');
    else
      raise EFileTypeError.Create('Unsupported file version.');
    end;
  finally
    f.Free;
  end;
end;

{$IFDEF SUPPORT_VER2}
procedure SynthLoadFromStream2(var Song:TSynthSong; Stream:TStream);
var i,j:integer;
begin
  Stream.Read(song.Creator,Sizeof(song.Creator));
  Stream.Read(song.Title,Sizeof(song.Title));
  Stream.Read(song.Composer,Sizeof(song.Composer));
  Stream.Read(song.TrackCount,Sizeof(Integer));
  SetLength(Song.Tracks,Song.TrackCount);
  for i:= 0 to Song.TrackCount-1 do begin
    Stream.Read(song.Tracks[i].TrackName,Sizeof(song.Tracks[j].TrackName));
    Stream.Read(song.Tracks[i].InstrumentID,Sizeof(song.Tracks[j].InstrumentID));
    Stream.Read(song.Tracks[i].EventCount,Sizeof(song.Tracks[j].EventCount));
    SetLength(Song.Tracks[i].Events,Song.Tracks[i].EventCount);
    for j:= 0 to Song.Tracks[i].EventCount-1 do begin
      Stream.Read(Song.Tracks[i].Events[j].TimeBefore,sizeof(Word));
      Stream.Read(Song.Tracks[i].Events[j].EventType,sizeof(TEventType));
      case Song.Tracks[i].Events[j].EventType  of
        seNone: ;
        seOn : ;
        seOff: ;
        sePitch: Stream.Read(Song.Tracks[i].Events[j].Freq,Sizeof(Single));
        sePan: Stream.Read(Song.Tracks[i].Events[j].Panning,Sizeof(Single));
        seRepeat: begin
          Stream.Read(Song.Tracks[i].Events[j].Last,Sizeof(Word));
          Stream.Read(Song.Tracks[i].Events[j].Times,Sizeof(Word));
        end;
        seText: begin
          Stream.Read(Song.Tracks[i].Events[j].Text[1],Sizeof(String25));
          TerminateString(Song.Tracks[i].Events[j].Text);
        end;
        seVolume: Stream.Read(Song.Tracks[i].Events[j].Volume,Sizeof(Single));
      end;  
    end;
 end;
end;
{$ENDIF}

{$IFDEF SUPPORT_VER1}
procedure SynthLoadFromStream1(var Song:TSynthSong; Stream:TStream);
var i,j,t:integer;
begin
  Stream.Read(song.Creator,Sizeof(song.Creator));
  Stream.Read(song.Title,Sizeof(song.Title));
  Stream.Read(song.Composer,Sizeof(song.Composer));
  Stream.Read(song.TrackCount,Sizeof(Integer));
  SetLength(Song.Tracks,Song.TrackCount);
  for i:= 0 to Song.TrackCount-1 do begin
    Stream.Read(song.Tracks[i].TrackName,Sizeof(song.Tracks[j].TrackName));
    Stream.Read(song.Tracks[i].InstrumentID,Sizeof(song.Tracks[j].InstrumentID));
    Stream.Read(song.Tracks[i].EventCount,Sizeof(song.Tracks[j].EventCount));
    SetLength(Song.Tracks[i].Events,Song.Tracks[i].EventCount);
    for j:= 0 to Song.Tracks[i].EventCount-1 do begin
      Stream.Read(t,sizeof(Integer));
      Song.Tracks[i].Events[j].TimeBefore:= t;
      Stream.Read(Song.Tracks[i].Events[j].EventType,sizeof(TEventType));
      case Song.Tracks[i].Events[j].EventType  of
        seNone: ;
        seOn : ;
        seOff: ;
        sePitch: Stream.Read(Song.Tracks[i].Events[j].Freq,Sizeof(Single));
        sePan: Stream.Read(Song.Tracks[i].Events[j].Panning,Sizeof(Single));
        seRepeat: begin
          Stream.Read(Song.Tracks[i].Events[j].Last,Sizeof(Word));
          Stream.Read(Song.Tracks[i].Events[j].Times,Sizeof(Word));
        end;
        seText: begin
          Stream.Read(Song.Tracks[i].Events[j].Text[1],Sizeof(String25));
          TerminateString(Song.Tracks[i].Events[j].Text);
        end;
        seVolume: Stream.Read(Song.Tracks[i].Events[j].Volume,Sizeof(Single));
      end;  
    end;
 end;
end;
{$ENDIF}

{$IFDEF SUPPORT_VER0}
procedure SynthLoadFromStream0(var Song:TSynthSong; Stream:TStream);
var i,j,t:integer;
begin
  Stream.Read(song.Creator,Sizeof(song.Creator));
  Stream.Read(song.Title,Sizeof(song.Title));
  Stream.Read(song.Composer,Sizeof(song.Composer));
  Stream.Read(song.TrackCount,Sizeof(Integer));
  SetLength(Song.Tracks,Song.TrackCount);
  for i:= 0 to Song.TrackCount-1 do begin
    Stream.Read(song.Tracks[i].TrackName,Sizeof(song.Tracks[j].TrackName));
    Stream.Read(song.Tracks[i].InstrumentID,Sizeof(song.Tracks[j].InstrumentID));
    Stream.Read(song.Tracks[i].EventCount,Sizeof(song.Tracks[j].EventCount));
    SetLength(Song.Tracks[i].Events,Song.Tracks[i].EventCount);
    for j:= 0 to Song.Tracks[i].EventCount-1 do begin
      Stream.Read(t,sizeof(Integer));
      Song.Tracks[i].Events[j].TimeBefore:= t;
      Stream.Read(Song.Tracks[i].Events[j].EventType,sizeof(TSynthEvent)-21-2);
      if Song.Tracks[i].Events[j].EventType=seText then begin
        Stream.Seek(-4,soFromCurrent);
        Stream.Read(Song.Tracks[i].Events[j].Text,Sizeof(String25));
      end;
    end;
 end;
end;
{$ENDIF}

initialization
finalization
  SynthStopSong;

end.
