{-----------------------------------------------------------------------------
 Unit Name: SynthUtils
 Author:    Sebastian Htter
 Date:      2006-10-29
 Purpose:   Simple wave functions, envelopes and harmonics
            Parts taken from a Tutorial by Nico Michaelis

 History:   2006-10-29 initial release
 
------------------------------------------------------------------------------
 Copyright Notice: If you copy or modify this file, this block
   _must_ be preserved! If you are using this code in your projects,
   I would like to see a word of thanks in the About-Box or a similar place.
-----------------------------------------------------------------------------}

unit SynthUtils;

interface

function Harmonics(x, f : Single) : Single;
function TriangleHarmonics(x, f : Single) : Single;
function SawtoothHarmonics(x, f : Single) : Single;
function SubHarminics(x, f : Single) : Single;

function Sgn(x : Single) : Single;

function WaveTriangle(x: Single): Single;
function WavePulse(x, Breite: Single): Single;
function WaveSawtooth(x: Single): Single;
function WaveSine(x: Single): Single;

function ADSREnvelope(t, a, d, s, r, maxlevel, holdlevel : Single): Single;
function ADEnvelopeCutOff(t, a, d, dmax, r, maxlevel, holdlevel : Single): Single;
function ASREnvelope(t, a, s, r, maxlevel : Single): Single;

implementation

uses
  Math;

const
  e = 2.71828182846;

function Harmonics(x, f : Single) : Single;
{Erzeugt aus Sini ein obertonreiches, normiertes Gesamt.
 x : Zeitlauf; f : Basis-Frequenz des zu erzeugenden Tones}
begin
  x := x * 2*Pi * f;
  Result :=          20 * Sin(x    );
  Result := Result + 13 * Sin(x * 2);
  Result := Result + 9  * Sin(x * 4);
  Result := Result + 5  * Sin(x * 8);
  Result := 1/(20+13+9+5) * Result;
end;

function TriangleHarmonics(x, f : Single) : Single;
{ Wie Obertoene, nur mit Dreiecken statt Sini }
begin
  x := x * f;
  Result :=          20 * WaveTriangle(x    );
  Result := Result + 13 * WaveTriangle(x * 2);
  Result := Result + 9  * WaveTriangle(x * 4);
  Result := Result + 5  * WaveTriangle(x * 8);
  Result := 1/(20+13+9+5) * Result;
end;

function SawtoothHarmonics(x, f : Single) : Single;
{ Wie Obertoene, nur mit Sgezhnen statt Sini }
begin
  x := x * f;
  Result :=          20 * WaveSawtooth(x    );
  Result := Result + 13 * WaveSawtooth(x * 2);
  Result := Result + 9  * WaveSawtooth(x * 4);
  Result := Result + 5  * WaveSawtooth(x * 8);
  Result := 1/(20+13+9+5) * Result;
end;

function SubHarminics(x, f : Single) : Single;
{ Wie Obertoene, nur Untertoene statt Obertoene }
begin
  x := x * 2*Pi * f;
  Result :=          20 * Sin(x    );
  Result := Result + 13 * Sin(1/2 * x);
  Result := Result + 9  * Sin(1/4 * x);
  Result := Result + 5  * Sin(1/8 * x);
  Result := 1/(20+13+9+5) * Result;
end;




function Sgn(x : Single) : Single;
begin
  if x < 0 then
    Result := -1.0
  else
    Result := +1.0;
end;

function WaveTriangle(x: Single): Single;
begin
  Result := Abs((frac(x) - 0.5) * 2*2)-1;
end;

function WavePulse(x, Breite: Single): Single;
begin
  Result := Sgn(frac(x) - Breite);
end;

function WaveSawtooth(x: Single): Single;
begin
  Result := (-1 + 2 * frac(x));
end;

function WaveSine(x: Single): Single;
begin
  Result := Sin(2*Pi*x);
end;

function ADSREnvelope(t, a, d, s, r, maxlevel, holdlevel : Single): Single;
const
  Start = -4.65;
  Intervallaenge = 1.0 - Start;

var
  Pos, x, delta : Single;
begin
  if t < 0 then
  begin
    Result := 0;
    Exit;
  end;

  //Attack
  Pos := a;
  if t < Pos then
  begin
    x := 1.0 - t/a; //x zwischen 0 und 1
    x := x * Intervallaenge + Start; //x auf das Intervall mappen
    Result := (1.0 - 1/e*(Power(e, x)))*maxlevel;
    Exit;
  end;

  //Decay
  Pos := Pos + d;
  if t <= Pos then
  begin
    delta := maxlevel - holdlevel;
    if delta <= 0.0 then
    begin
      Result := holdlevel;
      Exit
    end;
    x := 1.0 - (t - Pos + d)/d;
    x := x * Intervallaenge + Start;
    Result := Maxlevel - (1.0 - 1/e*(Power(e, x)))*delta;
    Exit;
  end;

  //Sustain
  Pos := Pos + s;
  if t <= Pos then
  begin
    Result := holdlevel;
    Exit
  end;

  //Release
  Pos := Pos + r;
  if t <= Pos then
  begin
    x := 1.0 - (t - Pos + r)/r;
    x := x * Intervallaenge + Start;
    Result := holdlevel - holdlevel * (1.0 - 1/e*(Power(e, x)));
    Exit;
  end;

  Result := 0.0;
end;

function ADEnvelopeCutOff(t, a, d, dmax, r, maxlevel, holdlevel : Single): Single;
const
  Start = -4.65;
  Intervallaenge = 1.0 - Start;

var
  Pos, x, delta : Single;
begin
  if t < 0 then
  begin
    Result := 0;
    Exit;
  end;
  if d > dmax then
    d := dmax;

  //Attack
  Pos := a;
  if t < Pos then
  begin
    x := 1.0 - t/a; //x zwischen 0 und 1
    x := x * Intervallaenge + Start; //x auf das Intervall mappen
    Result := (1.0 - 1/e*(Power(e, x)))*maxlevel;
    Exit;
  end;

  //Decay
  Pos := Pos + d;
  if t <= Pos then
  begin
    delta := maxlevel - holdlevel;
    if delta <= 0.0 then
    begin
      Result := holdlevel;
      Exit
    end;
    x := 1.0 - (t - Pos + d)/dmax;
    x := x * Intervallaenge + Start;
    Result := Maxlevel - (1.0 - 1/e*(Power(e, x)))*delta;
    Exit;
  end;

  begin
    delta := maxlevel - holdlevel;
    if delta <= 0.0 then
    begin
      Result := holdlevel;
      Exit
    end;
    x := 1.0 - d/dmax;
    x := x * Intervallaenge + Start;
    holdlevel := Maxlevel - (1.0 - 1/e*(Power(e, x)))*delta;
  end;

  //Release
  Pos := Pos + r;
  if t <= Pos then
  begin
    x := 1.0 - (t - Pos + r)/r;
    x := x * Intervallaenge + Start;
    Result := holdlevel - holdlevel * (1.0 - 1/e*(Power(e, x)));
    Exit;
  end;

  Result := 0.0;
end;

function ASREnvelope(t, a, s, r, maxlevel : Single): Single;
const
  Start = -4.65;
  Intervallaenge = 1.0 - Start;

var
  Pos, x : Single;
begin
  if t < 0 then
  begin
    Result := 0;
    Exit;
  end;

  //Attack
  Pos := a;
  if t < Pos then
  begin
    x := 1.0 - t/a; //x zwischen 0 und 1
    x := x * Intervallaenge + Start; //x auf das Intervall mappen
    Result := (1.0 - 1/e*(Power(e, x)))*maxlevel;
    Exit;
  end;

  //Sustain
  Pos := Pos + s;
  if t <= Pos then
  begin
    Result := maxlevel;
    Exit
  end;

  //Release
  Pos := Pos + r;
  if t <= Pos then
  begin
    x := 1.0 - (t - Pos + r)/r;
    x := x * Intervallaenge + Start;
    Result := maxlevel - maxlevel * (1.0 - 1/e*(Power(e, x)));
    Exit;
  end;

  Result := 0.0;
end;



end.
 